/* Copyright (c) 2018
   Charles Jordan (Hokkaido University, Sapporo, Japan),
   Michael Joswig, Lars Kastner (Technische Universitaet Berlin, Germany)

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any
   later version: http://www.gnu.org/licenses/gpl.txt.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
-----------------------------------------------------------------------------
*/

///////////////////////////////////////////////////////////////////////////////
// SymmetryBlackBox.hh
// 
// This file helps us deal with the symmetry group. It contains methods to
// check equivalence of triangulations, compute a canonical representative,
// etc.
//
// Furthermore it contains caching mechanisms for some of the computations.
///////////////////////////////////////////////////////////////////////////////
#ifndef SYMMETRYBLACKBOX_HH
#define SYMMETRYBLACKBOX_HH
#include "TriangNode.hh"
#include "Symmetry.hh"
#include "LRUCache.hh"
#include "GKZ.hh"
#include "mptopcom_typedefs.hh"
#include "GroupDecomposer.hh"
#include "TriangulationOptimizer.hh"

#include "polymake/group/permlib_tools.h"

template<typename checkFine, typename smallGroup, typename lexMinimal>
class SymmetryBlackBox {
   private:
      mutable Int current=0;
      mutable LRUCache<TriangNode, TriangNode> canonicalCache;
      mutable LRUCache<TriangNode, gkz_vector_type> canonicalGKZCache;
      mutable LRUCache<TriangNode, Int> classCache;
      const SymmetryGroup& group;
      const gkz_type& gkz;
      const GroupDecomposer gd;

   public:
      SymmetryBlackBox(Int size, const SymmetryGroup& gp, const gkz_type& gkz_in):
         canonicalGKZCache(size), canonicalCache(size),
         classCache(size), group(gp), gkz(gkz_in), gd(gp, gkz_in)
         { }

      bool isTrivial() const{
         return group.size() == 0;
      }

      TriangNode get_canonical_representative_inner(const TriangNode& t) const{
         if(group.card() == 0){
            return t;
         } else {
            using Optimizer = typename std::conditional<smallGroup::value, FullOrbitTriangulationOptimizer<false, lexMinimal, checkFine>,  TriangulationOptimizer<false, lexMinimal>>::type;
            Optimizer td(t, gkz.evaluate(t), gd, gkz);
            TriangNode result(td.get_optimal());
            if(!canonicalGKZCache.exist(result)){
               canonicalGKZCache.put(result, td.get_optimal_gkz());
            }
            return result;
         }
      }

      Int get_orbit_size(const TriangNode& t) const{
         pm::Set<TriangNode> eqclass;
         eqclass += t;
         for(SymmetryGroup::const_iterator sym_iter = group.begin();
               sym_iter != group.end();
               ++sym_iter){
            Symmetry g(*sym_iter);
            eqclass += (g(t));
         }
         return eqclass.size();
      }


      const TriangNode& get_canonical_representative(const TriangNode& t) const{
         if(!canonicalCache.exist(t)){
            TriangNode result = get_canonical_representative_inner(t);
            canonicalCache.put(t, result);
         }
         return canonicalCache.get(t);
      }

      const gkz_vector_type& get_canonical_gkz(const TriangNode& t) const{
         const TriangNode& key(get_canonical_representative(t));
         return get_gkz_of_canonical(key);
      }
      
      // Bypass computation of representative
      const gkz_vector_type& get_gkz_of_canonical(const TriangNode& canonical) const{
         if(!canonicalGKZCache.exist(canonical)){
            gkz_vector_type result(gkz.evaluate(canonical));
            canonicalGKZCache.put(canonical, result);
         }
         return canonicalGKZCache.get(canonical);
      }

      // Check equivalence by checking equality of the canonical representative.
      bool are_equivalent_triangulations(const TriangNode& t1, const TriangNode& t2) const{
         return get_canonical_representative(t1) == get_canonical_representative(t2);
      }


      Int getClassNumber(const TriangNode& t) const{
         const TriangNode& canonical = get_canonical_representative(t);
         if(classCache.exist(canonical)){
            return classCache.get(canonical);
         } else {
            classCache.put(canonical, current);
            current++;
            return classCache.get(canonical);
         }
      }

      void put(const TriangNode& t){
         classCache.put(t, current);
         current++;
      }
      
};
#endif
